#include "jserver.h"

/* mod_offline must go before mod_presence */

/* get the user's offline options */
xmlnode mod_offline_get(udata u)
{
    xmlnode ret;
 
    log_debug("mod_offline","getting %s's offline options",u->user);
 
    /* get the existing options */
    ret = js_xdb_get(u, NS_OFFLINE);
    if(ret == NULL)
    {
        log_debug("mod_offline","creating options container");
        ret = xmlnode_new_tag("offline");
        xmlnode_put_attrib(ret,"xmlns",NS_OFFLINE);
        js_xdb_set(u,NS_OFFLINE,ret);
    }
 
    return ret;
}

/* util to get the offline options xml data for this user/resource */
xmlnode mod_offline_get_res(xmlnode opts, char *res)
{
    xmlnode ret;
    char *query;

    if(opts == NULL || res == NULL) return NULL;

    log_debug("mod_offline","fetching %s options",res);

    /* query for any existing item */
    query = malloc(strlen(res) + 7);
    *query = '\0';
    strcat(query,"x?res=");
    strcat(query,res);
    ret = xmlnode_get_tag(opts, query);
    free(query);

    return ret;
}

/* handle an offline message */
mreturn mod_offline_message(mapi m)
{
    xmlnode opts, msgopts;
    char *forward;
    session top;

    log_debug("mod_offline","handling message for %s",m->user->user);

    /* get user offline options */
    opts = mod_offline_get(m->user);

    /* check and see if this resource has any special options */
    msgopts = xmlnode_get_tag(mod_offline_get_res(opts, m->packet->to->resource),"messages");

    if(msgopts != NULL)
    {
        if((forward = xmlnode_get_attrib(msgopts,"to")) != NULL)
        { /* this resource has messages being forwarded */
            xmlnode_put_attrib(m->packet->x,"to",forward);
            jpacket_reset(m->packet);
            js_deliver(m->packet);
        }else{
            /* store just for this resource */
            js_delay(m->packet->x,"Offline Storage");
            xmlnode_insert_tag_node(msgopts,m->packet->x);
            js_xdb_set(m->user,NS_OFFLINE,opts);
            xmlnode_free(m->packet->x);
        }
    }else if((top = js_session_primary(m->user)) != NULL){
        /* there's an existing session, just give it to them */
        js_psend(top->mp_to,m->packet);
    }else{
        /* alone am I */
        js_delay(m->packet->x,"Offline Storage");
        xmlnode_insert_tag_node(opts,m->packet->x);
        js_xdb_set(m->user,NS_OFFLINE,opts);
        xmlnode_free(m->packet->x);
    }

    return M_HANDLED;
}

mreturn mod_offline_presence(mapi m)
{
    xmlnode opts, cur, pres;

    if(jpacket_subtype(m->packet) != JPACKET__PROBE) return M_PASS;

    log_debug("mod_offline","handling presence probe for %s",m->user->user);

    /* get user offline options */
    opts = mod_offline_get(m->user);

    /* loop through the resources */
    for(cur = xmlnode_get_firstchild(opts); cur != NULL; cur = xmlnode_get_nextsibling(cur))
    {
        pres = xmlnode_get_tag(cur,"presence");
        if(pres == NULL || js_session_get(m->user,xmlnode_get_attrib(cur,"res")) != NULL) continue;

        /* there's some offline presence, return it to the prober */
        pres = xmlnode_dup(pres);
        xmlnode_put_attrib(pres,"to",jid_full(m->packet->from));
        log_debug("mod_offline","sending offline presence %s",xmlnode2str(pres));
        js_deliver(jpacket_new(pres));
    }
    
    return M_PASS;
}

/* just breaks out to our message/presence offline handlers */
mreturn mod_offline_handler(mapi m, void *arg)
{
    if(m->packet->type == JPACKET_MESSAGE) return mod_offline_message(m);
    if(m->packet->type == JPACKET_PRESENCE) return mod_offline_presence(m);

    return M_IGNORE;
}

/* filters the unavailables sent by the user for the NS_OFFLINE extension */
void mod_offline_out_unavailable(mapi m, void *arg)
{
    xmlnode *x = (xmlnode *)arg;
    xmlnode newx;

    newx = xmlnode_get_tag(m->packet->x,"x?xmlns=" NS_OFFLINE);
    if(newx != NULL)
    { /* store the offline opts for later permanence, and make sure nobody else sees it */
        log_debug("mod_offline","filtering offline options");
        xmlnode_hide(newx);
        xmlnode_free(*x);
        *x = xmlnode_dup(newx);
    }
}

/* watches for when the user is available and sends out offline messages */
void mod_offline_out_available(mapi m)
{
    xmlnode opts, resopts, cur;

    log_debug("mod_offline","avability established, check for messages");

    opts = mod_offline_get(m->user);
    resopts = mod_offline_get_res(opts,m->s->res);

    /* check for resource ones */
    for(cur = xmlnode_get_firstchild(xmlnode_get_tag(resopts,"messages")); cur != NULL; cur = xmlnode_get_nextsibling(cur))
    {
        if(j_strcmp(xmlnode_get_name(cur),"message") != 0) continue;

        js_psend(m->s->mp_to,jpacket_new(xmlnode_dup(cur)));
        xmlnode_hide(cur);
    }

    /* check for global ones */
    for(cur = xmlnode_get_firstchild(opts); cur != NULL; cur = xmlnode_get_nextsibling(cur))
    {
        if(j_strcmp(xmlnode_get_name(cur),"message") != 0) continue;

        js_psend(m->s->mp_to,jpacket_new(xmlnode_dup(cur)));
        xmlnode_hide(cur);
    }

    /* messages are gone, save the new sun-dried opts container */
    js_xdb_set(m->user, NS_OFFLINE, opts);
}

mreturn mod_offline_out(mapi m, void *arg)
{
    if(m->packet->type != JPACKET_PRESENCE) return M_IGNORE;

    if(jpacket_subtype(m->packet) == JPACKET__UNAVAILABLE)
        mod_offline_out_unavailable(m, arg);
    if(jpacket_subtype(m->packet) == JPACKET__AVAILABLE && m->s->priority < 0)
        mod_offline_out_available(m);

    return M_PASS;
}

/* stores the offline options */
mreturn mod_offline_end(mapi m, void *arg)
{
    xmlnode *x = (xmlnode *)arg;
    xmlnode opts, resopts, pres, cur;

    if(*x == NULL) return M_PASS;

    log_debug("mod_offline","storing offline options %s",xmlnode2str(*x));

    /* store the offline presence */
    pres = xmlnode_get_tag(*x, "presence");
    if(pres != NULL)
    {
        xmlnode_hide(pres);
        /* insert the current session presence (which is always unavailable) and set the from attrib for convenience */
        xmlnode_put_attrib(xmlnode_insert_tag_node(*x,m->s->presence),"from",jid_full(m->s->id));
    }

    /* get the container */
    opts = mod_offline_get(m->user);

    /* ditch the old options */
    resopts = mod_offline_get_res(opts,m->s->res);
    xmlnode_hide(resopts);

    /* store the new options */
    xmlnode_put_attrib(xmlnode_insert_tag_node(opts,*x),"res",m->s->res);
    js_xdb_set(m->user,NS_OFFLINE,opts);
    xmlnode_free(*x);

    /* if there were any old offline messages stored (shouldn't be, but this is a strange world [eerie-noises] ) */
    for(cur = xmlnode_get_firstchild(xmlnode_get_tag(resopts,"messages")); cur != NULL; cur = xmlnode_get_nextsibling(cur))
    {
        if(j_strcmp(xmlnode_get_name(cur),"message") != 0) continue;

        /* redeliver stored messages */
        js_deliver(jpacket_new(xmlnode_dup(cur)));
    }

    return M_PASS;
}

/* sets up the per-session listeners */
mreturn mod_offline_session(mapi m, void *arg)
{
    xmlnode *x;

    log_debug(ZONE,"session init");

    /* this holds the last known offline setting */
    x = pmalloc(m->s->p, sizeof(xmlnode *));
    *x = NULL;

    js_mapi_session(PS_OUT, m->s, mod_offline_out, x);
    js_mapi_session(PS_END, m->s, mod_offline_end, x);

    return M_PASS;
}

void mod_offline(void)
{
    log_debug("mod_offline","init");
    js_mapi_register(P_OFFLINE, mod_offline_handler, NULL);
    js_mapi_register(P_SESSION, mod_offline_session, NULL);
}

