#include "jserver.h"


session js_session_new(jid owner, session_onSend send, void *arg)
{
    /* creats a new session, registers the resource for it, and spawns a thread to handle it */
    thread t;
    session s;
    pth_attr_t attr;
    jid uid;

    if(owner == NULL || owner->resource == NULL) /* illegal */
        return NULL;
 
    log_debug(ZONE,"session_create");
 
    /* create session thread */
    t = thread_new();
    s = pmalloc(t->p, sizeof(struct session_struct));
    t->data = (void *)s;
    s->t = t;
    s->p = t->p;

    /* setup internal communication points */
    s->mp_from = pth_msgport_create("js_session_from");
    s->mp_to = pth_msgport_create("js_session_to");
    s->m_out = s->m_in = s->m_end = NULL;

    /* save the send callback for the service */
    s->send = send;
    s->arg = arg;

    /* session identity */
    s->id = jid_new(t->p, jid_full(owner));
    uid = jid_new(t->p, jid_full(owner));
    jid_set(uid, NULL, JID_RESOURCE);
    s->uid = uid;
    s->res = pstrdup(s->p, owner->resource);
    s->u = js_user(owner->user);

    /* default settings */
    s->exit_flag = 0;
    s->roster = 0;
    s->priority = -1;
    s->presence = NULL;
    s->c_in = s->c_out = 0;

    /* remove any other session w/ this resource */
    js_session_end(js_session_get(s->u, owner->resource), "Replaced by new connection");

    /* make sure we're linked with the user */
    s->next = s->u->sessions;
    s->u->sessions = s;
    s->u->scount++;

    /* start thread */
    attr = pth_attr_new();
    pth_attr_set(attr, PTH_ATTR_JOINABLE, FALSE);
    pth_attr_set(attr, PTH_ATTR_STACK_SIZE, etherx_stack_default);
    t->id = pth_spawn(attr, js_session_main, (void *)t);
    pth_attr_destroy(attr);
    pth_yield(NULL); /* let it have a chance to start up */
 
    return s;
}

void js_session_end(session s, char *reason)
{
    xmlnode x;
    jid to_id;
    session cur;

    if(s == NULL || s->exit_flag == 1 || reason == NULL)
        return;

    log_debug(ZONE,"end %d '%s'",s,reason);

    /* flag the session to exit ASAP */
    s->exit_flag = 1;

    /* make sure we're not the primary session */
    s->priority = -1;

    /* create an unavailable presence packet for this action if needed */
    x = xmlnode_new_tag("presence");
    xmlnode_put_attrib(x,"type","unavailable");
    xmlnode_put_attrib(x,"from",jid_full(s->id));
    xmlnode_insert_cdata(xmlnode_insert_tag(x,"status"),reason,strlen(reason));

    /* if the last known presence was available, update it */
    if(s->presence != NULL && j_strcmp(xmlnode_get_attrib(s->presence, "type"), "unavailable") != 0)
    {
        xmlnode_free(s->presence);
        s->presence = xmlnode_dup(x);
    }

    /* unlist ourselves */
    if(s == s->u->sessions)
    {
        s->u->sessions = s->next;
    }else{
        for(cur = s->u->sessions; cur->next != s; cur = cur->next);
        cur->next = s->next;
    }
    s->u->ref++; /* so it doesn't get freed */

    /* give the session some CPU */
    js_psend(s->mp_from,jpacket_new(x));
}

void *js_session_main(void *arg)
{
    mmaster master;
    thread t = (thread)arg;
    session s = (session)(t->data);
    pth_event_t from, to, ring;
    jpq q;

    log_debug(ZONE,"THREAD:SESSION starting");

    from = pth_event(PTH_EVENT_MSG,s->mp_from);
    to = pth_event(PTH_EVENT_MSG,s->mp_to);
    ring = pth_event_concat(from, to, NULL);

    master = js_mapi_master(P_SESSION);
    js_mapi_call(P_SESSION, master->l, NULL, s->u, s, NULL);

    while(1)
    {
        log_debug(ZONE,"SESSION(%s)->pth",s->u->user);
        pth_wait(ring);
        log_debug(ZONE,"pth->SESSION(%s)",s->u->user);

        if(s->exit_flag)
            break;

        while(1)
        {
            /* this loop looks for messages from our user that need to be delivered elsewhere */
            q = (jpq)pth_msgport_get(s->mp_from);
            if(q == NULL)
                break;

            /* at least we must have a valid packet */
            if(q->p->type == JPACKET_UNKNOWN)
            {
                js_error(q->p->x,TERROR_BAD);
                jpacket_reset(q->p);
                js_psend(s->mp_to,q->p);
                return;
            }

            log_debug(ZONE,"THREAD:SESSION:FROM received a message!");
            s->c_out++;

            /* make sure we have our from set correctly for outgoing crap */
            xmlnode_put_attrib(q->p->x,"from",jid_full(s->id));
            q->p->from = jid_new(q->p->p,jid_full(s->id));

            /* let the modules have their heyday */
            if(js_mapi_call(PS_OUT, s->m_out, q->p, s->u, s, jpacket_subtype(q->p)))
                continue;

            if(q->p->to == NULL)
            {
                xmlnode_put_attrib(q->p->x,"to",q->p->from->server);
                q->p->to = jid_new(q->p->p,q->p->from->server);
            }

            /* pass these to the general delivery function */
            js_deliver(q->p);
        }

        if(s->exit_flag)
            break;

        while(1)
        {
            /* this loop looks for messages to our user that need to be sent to them */
            q = (jpq)pth_msgport_get(s->mp_to);
            if(q == NULL)
                break;

            log_debug(ZONE,"THREAD:SESSION:TO received data from %s!",jid_full(q->p->from));
            s->c_in++;

            /* let the modules have their heyday */
            if(js_mapi_call(PS_IN, s->m_in, q->p, s->u, s, jpacket_subtype(q->p)))
                continue;

            /* we need to check again, s->exit_flag *could* have changed within the modules at some point */
            if(s->exit_flag)
            { /* save it if it was a message, and sk'daddle */
                if(q->p->type == JPACKET_MESSAGE)
                    js_deliver(q->p);
                else
                    xmlnode_free(q->p->x);
                break;
            }

            /* deliver outgoing for this session to the onSend event passed when this session was created */
            (s->send)(s, q->p, s->arg);
        }

    }

    log_debug(ZONE,"THREAD:SESSION exiting");

    s->u->scount--;

    /* make sure the service knows the session is gone */
    (s->send)(s, NULL, s->arg);

    /* let the modules have their heyday */
    js_mapi_call(PS_END, s->m_end, NULL, s->u, s, 0);

    xmlnode_free(s->presence);
    pth_event_free(ring,PTH_FREE_ALL);

    /* the to be processed packets, if any */
    while(1)
    {
        q = (jpq)pth_msgport_get(s->mp_from);
        if(q == NULL)
            break;
        xmlnode_free(q->p->x);
    }
    pth_msgport_destroy(s->mp_from);

    /* the delivery waiting packets, if any */
    while(1)
    {
        q = (jpq)pth_msgport_get(s->mp_to);
        if(q == NULL)
            break;
        if(q->p->type == JPACKET_MESSAGE)
            js_deliver(q->p); /* re-deliver the messages only */
        else
            xmlnode_free(q->p->x);
    }
    pth_msgport_destroy(s->mp_to);

    s->u->ref--; /* let the user struct go  */
    thread_free(t);

    return NULL;
}

session js_session_get(udata user, char *res)
{
    session cur;
 
    if(user == NULL || res == NULL)
        return NULL;
 
    for(cur = user->sessions; cur != NULL; cur = cur->next)
        if(j_strcmp(res, cur->res) == 0)
            return cur;
 
    return NULL;
}

session js_session_primary(udata user)
{
    session cur, top;
 
    if(user == NULL || user->sessions == NULL)
        return NULL;
 
    /* find primary session */
    top = user->sessions;
    for(cur = top; cur != NULL; cur = cur->next)
        if(cur->priority > top->priority)
            top = cur;
 
    if(top->priority >= 0)
        return top;
    else
        return NULL;
}

