#include "svc_irc.h"

/* OUT: IRC --> Jabber XML */

/* start a new session for this connection */
void svc_irc_session(ircdata d)
{
    jpacket p;

    /* set up the jabber id and start the session */
    d->id = jid_new(d->t->p,d->server);
    jid_set(d->id, d->nick, JID_USER);
    jid_set(d->id, d->res, JID_RESOURCE);
    d->s = js_session_new(d->id,svc_irc_in,(void *)d);

    /* invite them to the main room */
    svc_irc_line(d,NULL,"INVITE",ROOM,1,d->nick);
    d->phase = PHASE_PRE;
}

void svc_irc_help_room(ircdata d)
{
    svc_irc_line(d,"nickserv","NOTICE","-------------------------------------------------------",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","Your roster is the users in this room, which you can manage with:",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","    /msg nickserv add|del|set user@server [nick]",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","You can manage your presence subscriptions with:",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","    /msg nickserv [un]subscribe[d] user@server [message]",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","Send messages to users using 'nick: message' or '/msg nick message'",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","Change presence by '/me on|off|away|xa|dnd status message' (you are currently 'on')",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","-------------------------------------------------------",1,d->nick);
}

/* create the room */
void svc_irc_room(ircdata d)
{
    jpacket p;

    /* send the new room info */
    svc_irc_line(d,d->nick,"JOIN",ROOM,0);
    svc_irc_line(d,NULL,"332",jid_full(d->id),2,d->nick,ROOM); /* title */

    /* get the roster */
    p = jpacket_new(js_iq_new(JPACKET__GET,NS_ROSTER));
    js_psend(d->s->mp_from,p);

    /* set the presence */
    p = jpacket_new(xmlnode_new_tag("presence"));
    js_psend(d->s->mp_from,p);

    svc_irc_line(d,NULL,"NOTICE","Welcome!",1,d->nick);
    svc_irc_help_room(d);

    d->phase = PHASE_NORM;
}

/* process a login, /msg login password [resource] */
int svc_irc_auth(ircdata d)
{
    jpacket p;
    int ret;

    p = jpacket_new(js_iq_new(JPACKET__SET,NS_AUTH));
    xmlnode_insert_cdata(xmlnode_insert_tag(p->iq, "username"),d->nick,-1);
    xmlnode_insert_cdata(xmlnode_insert_tag(p->iq, "password"),d->pass,-1);
    xmlnode_insert_cdata(xmlnode_insert_tag(p->iq, "resource"),d->res,-1);

    log_debug(ZONE,"auth %s",xmlnode2str(p->x));

    if(js_service_prescreen(p) == S_AUTHED)
    { /* start the session */
        xmlnode_free(p->x);
        return 1;
    }

    xmlnode_free(p->x);
    return 0;
}

/* process a registration, /msg register password  */
int svc_irc_reg(ircdata d)
{
    int ret;
    jpacket p;

    p = jpacket_new(js_iq_new(JPACKET__SET,NS_REGISTER));
    xmlnode_insert_cdata(xmlnode_insert_tag(p->iq, "username"),d->nick,-1);
    xmlnode_insert_cdata(xmlnode_insert_tag(p->iq, "password"),d->pass,-1);

    log_debug(ZONE,"register %s",xmlnode2str(p->x));
    js_service_prescreen(p);

    jpacket_reset(p);
    ret = jpacket_subtype(p);
    xmlnode_free(p->x);

    if(ret == JPACKET__RESULT)
        return 1;
    else
        return 0;
}

/* process session presence bit */
void svc_irc_pres(ircdata d, char *data)
{
    char *desc;
    xmlnode x, item;
    jid id;

    desc = strstr(data," ");
    if(desc != NULL)
    {
        *desc = '\0';
        ++desc;
    }

    if(strcmp(data,"off") == 0)
    {
        x = js_pres_new(JPACKET__UNAVAILABLE,NULL,desc);

        /* whack the presence cache */
        for(item = xmlnode_get_firstchild(d->roster); item != NULL; item = xmlnode_get_nextsibling(item))
        { /* make sure all the users are unavail */
            if(xmlnode_get_attrib(item,"jid") == NULL) continue;

            id = jid_new(xmlnode_pool(x),xmlnode_get_attrib(item,"jid"));
            if(ppdb_primary(d->pres,id) != NULL)
                svc_irc_line(d, "presence", "MODE", NULL, 3, ROOM, "-o", xmlnode_get_attrib(item,"name"));
        }
        ppdb_free(d->pres);
        d->pres = NULL;
    }else{
        x = js_pres_new(JPACKET__AVAILABLE,NULL,desc);
        if(strcmp(data,"on") != 0)
            xmlnode_insert_cdata(xmlnode_insert_tag(x,"show"),data,-1);
    }

    js_psend(d->s->mp_from,jpacket_new(x));
}

/* process a PRIVMSG */
void svc_irc_message(ircdata d, char *to, char *data)
{
    xmlnode x, body;
    jid id;
    jpacket p;
    char *colon, *space;
    int chat = 1;

    log_debug(ZONE,"message to %s data %X %s",to, *data, data);

    if(strcmp(to,ROOM) == 0)
    {
        colon = strstr(data,":");
        space = strstr(data," ");
        if(colon == NULL || (space != NULL && space < colon))
        {
            if(d->lastnick == NULL)
            {
                svc_irc_line(d,NULL,"NOTICE","You need to address someone specifically with nick:",1,d->nick);
                return;
            }
            to = d->lastnick;
        }else{
            *colon = '\0';
            colon++;
            to = data;
            data = colon;
            free(d->lastnick);
            d->lastnick = strdup(to);
        }
        chat = 0;
    }

    if(svc_irc_id(d,to) != NULL)
        to = svc_irc_id(d,to);

    x = xmlnode_new_tag("message");
    xmlnode_put_attrib(x, "to", to);
    if(chat)
        xmlnode_put_attrib(x,"type","chat");
    if(data != NULL)
    {
        body = xmlnode_insert_tag(x, "body");
        xmlnode_insert_cdata(body,data,strlen(data));
    }

    log_debug(ZONE,"sending %s",xmlnode2str(x));

    p = jpacket_new(x);
    js_psend(d->s->mp_from,p);
}

/* process roster requests, /msg nickserv add|del|set user@server [nick] */
void svc_irc_roster(ircdata d, char *data)
{
    char *ids, *nick;
    xmlnode item;
    jpacket p;
    jid id;

    ids = strstr(data," ");
    if(ids == NULL) return;
    *ids = '\0';
    ids++;

    nick = strstr(ids," ");
    if(nick != NULL)
    {
        *nick = '\0';
        nick++;
    }

    p = jpacket_new(js_iq_new(JPACKET__SET,NS_ROSTER));
    id = jid_new(p->p,ids);
    item = svc_irc_item(d,id);

    if(item == NULL)
    {
        item = xmlnode_insert_tag(p->iq,"item");
        xmlnode_put_attrib(item,"jid",ids);
    }else{
        item = xmlnode_insert_tag_node(p->iq,item);
    }

    if(nick != NULL)
        xmlnode_put_attrib(item,"name",nick);

    if(strcmp(data,"del") == 0)
        xmlnode_put_attrib(item,"subscription","remove");

    js_psend(d->s->mp_from,p);
}

/* process presence subscription requests, /msg nickserv [un]subscribe[d] user@server description... */
void svc_irc_s10n(ircdata d, char *data)
{
    char *id, *desc;
    xmlnode x;

    id = strstr(data," ");
    if(id == NULL) return;
    *id = '\0';
    id++;

    desc = strstr(id," ");
    if(desc != NULL)
    {
        *desc = '\0';
        desc++;
    }

    x = xmlnode_new_tag("presence");
    xmlnode_put_attrib(x, "to", id);
    xmlnode_put_attrib(x, "type", data);
    if(desc != NULL)
        xmlnode_insert_cdata(xmlnode_insert_tag(x,"status"),desc,strlen(desc));

    js_psend(d->s->mp_from,jpacket_new(x));
}

/* handle who queries */
void svc_irc_who(ircdata d, char *to)
{
    xmlnode cur;
    char *ids, *nick, *user, *server;
    pool p;
    jid id;

    log_debug(ZONE,"handling who query to %s",to);

    /* /who ROOM */
    if(j_strcmp(to,ROOM) == 0)
    {
        if(d->roster == NULL)
        {
            d->whoflag = 1;
            return;
        }
        p = pool_new();
        for(cur = xmlnode_get_firstchild(d->roster); cur != NULL; cur = xmlnode_get_nextsibling(cur))
        {
            if((ids = xmlnode_get_attrib(cur,"jid")) != NULL)
            {
                id = jid_new(p,ids);
                nick = svc_irc_nick(d, id);
                user = strstr(nick,"!");
                *user = '\0';
                ++user;
                server = strstr(user,"@");
                *server = '\0';
                ++server;
                svc_irc_line(d,NULL,"352",NULL,9,d->nick,ROOM,user,server,d->server,nick,"H",":0",ids);
            }
        }
        svc_irc_line(d,NULL,"315","End of /who list",2,d->nick,ROOM);
        pool_free(p);

        return;
    }

    /* /who user */

}

void svc_irc_help_auth(ircdata d)
{
    svc_irc_line(d,"nickserv","NOTICE","-------------------------------------------------------",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","You need to have an account to utilize this service.",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","Create a new account using your current nickname:",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","    /msg nickserv register password",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","Or authenticate your nickname with an existing account:",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","    /msg nickserv identify password [resource]",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","-------------------------------------------------------",1,d->nick);
}
void svc_irc_welcome(ircdata d)
{
    svc_irc_line(d,NULL,"001","Welcome to Jabber's IRC Service",1,d->nick);
    svc_irc_line(d,NULL,"004",NULL,5,d->nick,xmlnode_get_data(js_config("names/default")),"1.2.3","foo","bar");
}

void svc_irc_out(ircdata d, char *cmd, char *to, char *args, char *data)
{
    char *space, *space2;

    log_debug(ZONE,"command %s to %s with args %s and data %s",cmd,to,args,data);

    if(strcmp(cmd,"PONG") == 0)
        return;

    if(d->phase == PHASE_INIT)
    {
        if(strcmp(cmd,"USER") == 0 && to != NULL)
        {
            d->nick = pstrdup(d->t->p, to);
            svc_irc_welcome(d);
            svc_irc_help_auth(d);
            d->phase = PHASE_LOGIN;
        }
        return;
    }

    if(d->phase == PHASE_LOGIN)
    {
        if(strcmp(cmd,"NICK") == 0)
        {
            if(strlen(to) == 0) to = data; /* mirc does this?  grr */
            svc_irc_line(d,d->nick,"NICK",to,0);
            d->nick = pstrdup(d->t->p, to);
            return;
        }
        if(strcmp(cmd,"PRIVMSG") == 0 && to != NULL && strcasecmp(to,"nickserv") == 0 && data != NULL)
        {
            space = strstr(data," ");
            if(space == NULL)
            {
                svc_irc_line(d,"nickserv","NOTICE","Failed!",1,d->nick);
                return;
            }
            *space = '\0';
            ++space;
            space2 = strstr(space," ");
            if(space2 == NULL)
            {
                d->res = pstrdup(d->t->p,"irc");
            }else{
                *space2 = '\0';
                ++space2;
                d->res = pstrdup(d->t->p,space2);
            }
            d->pass = pstrdup(d->t->p, space);

            if(strcasecmp(data,"identify") == 0 && svc_irc_auth(d))
            {
            }else if(strcasecmp(data, "register") == 0 && svc_irc_reg(d) && svc_irc_auth(d)){
                svc_irc_line(d,"nickserv","NOTICE","Account Created",1,d->nick);
            }else{
                svc_irc_line(d,"nickserv","NOTICE","Failed!",1,d->nick);
                return;
            }
            svc_irc_session(d);
        }else{
            svc_irc_help_auth(d);
        }
        return;
    }

    /* before you're in the room */
    if(d->phase == PHASE_PRE)
    {
        if(strcmp(cmd,"JOIN") == 0 && j_strcmp(to, ROOM) == 0)
            svc_irc_room(d);
        else
            svc_irc_line(d,NULL,"NOTICE","You may only /join and /msg at this point.",1,d->nick);
        return;
    }

    /* deal w/ normal traffic here */

    if(strcmp(cmd,"PRIVMSG") == 0 && to != NULL && strcasecmp(to,"nickserv") == 0 && data != NULL)
    { /* /msg nickserv crap */
        if(strstr(data,"subscribe") != NULL)
            svc_irc_s10n(d,data);
        else
            svc_irc_roster(d,data);
        return;
    }

    if(strcmp(cmd,"PRIVMSG") == 0 && to != NULL &&  strcasecmp(to,"debug") == 0)
    { /* turn on debugging :) */
        d->debug ^= 1;
        return;
    }

    if(strcmp(cmd,"PRIVMSG") == 0 && *data == 1)
    { /* CTCP stuff */
        /* cur off the CTCP bits */
        data[strlen(data) - 1] = '\0';
        data++;

        if(strncmp(data,"ACTION ",7) == 0)
        {
            if(j_strcmp(to,ROOM) == 0)
            { /* /me in the room controls presence */
                svc_irc_pres(d,data + 7);
                return;
            }else{ /* make it a literal '/me' */
                data += 3;
                data[0] = '/';
                data[1] = 'm';
                data[2] = 'e';
            }
        }else{
            svc_irc_line(d,NULL,"NOTICE",NULL,3,d->nick,":I don't understand CTCP",data);
            return;
        }
    }

    if(strcmp(cmd,"PRIVMSG") == 0 && to != NULL)
    {
        svc_irc_message(d,to,data);
        return;
    }

    if(strcmp(cmd,"WHO") == 0 && to != NULL)
    {
        svc_irc_who(d,to);
        return;
    }

    if(strcmp(cmd,"MODE") == 0)
        return;

    svc_irc_line(d,NULL,"NOTICE","I don't understand!",1,d->nick);
}

