#include "jserver.h"

/* util for making presence packets */
xmlnode js_pres_new(int type, char *to, char *status)
{
    xmlnode pres;

    pres = xmlnode_new_tag("presence");
    switch(type)
    {
        case JPACKET__SUBSCRIBE:
            xmlnode_put_attrib(pres,"type","subscribe");
            break;
        case JPACKET__UNSUBSCRIBE:
            xmlnode_put_attrib(pres,"type","unsubscribe");
            break;
        case JPACKET__SUBSCRIBED:
            xmlnode_put_attrib(pres,"type","subscribed");
            break;
        case JPACKET__UNSUBSCRIBED:
            xmlnode_put_attrib(pres,"type","unsubscribed");
            break;
        case JPACKET__PROBE:
            xmlnode_put_attrib(pres,"type","probe");
            break;
        case JPACKET__UNAVAILABLE:
            xmlnode_put_attrib(pres,"type","unavailable");
            break;
    }
    if(to != NULL)
        xmlnode_put_attrib(pres,"to",to);
    if(status != NULL)
        xmlnode_insert_cdata(xmlnode_insert_tag(pres,"status"),status,strlen(status));

    return pres;
}

/* util for making IQ packets */
xmlnode js_iq_new(int type, char *ns)
{
    xmlnode iq;

    iq = xmlnode_new_tag("iq");
    switch(type)
    {
        case JPACKET__GET:
            xmlnode_put_attrib(iq,"type","get");
            break;
        case JPACKET__SET:
            xmlnode_put_attrib(iq,"type","set");
            break;
        case JPACKET__RESULT:
            xmlnode_put_attrib(iq,"type","result");
            break;
        case JPACKET__ERROR:
            xmlnode_put_attrib(iq,"type","error");
            break;
    }
    xmlnode_put_attrib(xmlnode_insert_tag(iq,"query"),"xmlns",ns);

    return iq;
}

/* returns the priority on a presence packet */
int js_priority(xmlnode x)
{
    char *str;
    int p;

    if(x == NULL)
        return -1;

    if(xmlnode_get_attrib(x,"type") != NULL)
        return -1;

    x = xmlnode_get_tag(x,"priority");
    if(x == NULL)
        return 0;

    str = xmlnode_get_data((x));
    if(str == NULL)
        return 0;

    p = atoi(str);
    if(p >= 0)
        return p;
    else
        return 0;
}

xmlnode js_iqreply(xmlnode x)
{
    char *to, *from;
    xmlnode cur;

    to = xmlnode_get_attrib(x,"to");
    from = xmlnode_get_attrib(x,"from");

    xmlnode_put_attrib(x,"type","result");

    if(to != NULL)
        xmlnode_put_attrib(x,"from",to);
    if(from != NULL)
        xmlnode_put_attrib(x,"to",from);

    /* hide all children of the iq, they go back empty */
    for(cur = xmlnode_get_firstchild(x); cur != NULL; cur = xmlnode_get_nextsibling(cur))
        xmlnode_hide(cur);

    return x;
}

char *js_timestamp(void)
{
	time_t t;
	struct tm *new_time;
	static char timestamp[18];
	int ret;
	
	t = time(NULL);

	if(t == (time_t)-1)
		return NULL;
	new_time = gmtime(&t);

	ret = snprintf(timestamp, 18, "%d%02d%02dT%02d:%02d:%02d", 1900+new_time->tm_year,
					new_time->tm_mon+1, new_time->tm_mday, new_time->tm_hour,
					new_time->tm_min, new_time->tm_sec);

	if(ret == -1)
		return NULL;

	return timestamp;
}

void js_error(xmlnode x, terror terr)
{
    char *to, *from;
    xmlnode err;
    char code[4];

    xmlnode_put_attrib(x,"type","error");
    err = xmlnode_insert_tag(x,"error");

    snprintf(code,4,"%d",terr.code);
    xmlnode_put_attrib(err,"code",code);
    if(terr.msg != NULL)
        xmlnode_insert_cdata(err,terr.msg,strlen(terr.msg));

    to = xmlnode_get_attrib(x,"to");
    from = xmlnode_get_attrib(x,"from");
    xmlnode_put_attrib(x,"from",to);
    xmlnode_put_attrib(x,"to",from);
}

void js_bounce(xmlnode x, terror terr)
{
    if(j_strcmp(xmlnode_get_name(x),"presence") == 0 && j_strcmp(xmlnode_get_attrib(x,"type"),"subscribe") == 0)
    { /* bounce subscriptions only */
        js_iqreply(x); /* hack alert, but it get's the job done */
        xmlnode_put_attrib(x,"type","unsubscribed");
        xmlnode_insert_cdata(xmlnode_insert_tag(x,"status"),terr.msg,-1);
        js_deliver(jpacket_new(x));
        return;
    }

    if(j_strcmp(xmlnode_get_name(x),"presence") == 0 || j_strcmp(xmlnode_get_attrib(x,"type"),"error") == 0)
    { /* presence packets are just dropped */
        log_debug(ZONE,"dropping %d packet %s",terr.code,xmlnode2str(x));
        xmlnode_free(x);
        return;
    }

    js_error(x, terr);
    js_deliver(jpacket_new(x));
}

void js_delay(xmlnode msg, char *reason)
{
    xmlnode delay;

    delay = xmlnode_insert_tag(msg,"x");
    xmlnode_put_attrib(delay,"xmlns",NS_DELAY);
    xmlnode_put_attrib(delay,"from",xmlnode_get_attrib(msg,"to"));
    xmlnode_put_attrib(delay,"stamp",js_timestamp());
    if(reason != NULL)
        xmlnode_insert_cdata(delay,reason,strlen(reason));
}
