#include "jserver.h"

/* To any xdb users, xdb is an xmlnode consumer.  
 * any data coming from the xdb is managed by the xdb, 
 * any data sent back to it (which can be the same data/pointers only modified)
 * is also then managed by xdb.
 * 
 * Typically, you get the data with xdb_get, make your changes, 
 * then call xdb_set or just forget about it and move on.
 */

/* Notes for xdb_* components 
 * First come first serve
 * All data returned in a get is aggressively cached,
 * which means the get will not be called again, only the set when it changes.
 * To prevent the caching, always make sure the (xdb)->cache bit is 1 (even in a set)
 */

xdb js_xdb_fetch(xdb cur, char *ns)
{
    for(;cur != NULL; cur = cur->next)
        if(j_strcmp(cur->ns, ns) == 0)
            return cur;
    return NULL;
}

xmlnode js_xdb(command cmd, udata user, char *ns, xmlnode data)
{
    static xlist l = NULL;
    static xdb global = NULL;
    xlist newl, i;
    xdb cur;

    if(cmd == C_INIT)
    {
        log_debug(ZONE,"registered new xdb");
        newl = malloc(sizeof(_xlist));
        newl->c = (xcall)user;
        newl->arg = (void *)ns;
        newl->next = l;
        l = newl;
        return NULL;
    }

    /* first get the entry from the current cache, or create a cache entry */
    if(user != NULL)
    {
        cur = js_xdb_fetch(user->x_cache, ns);
        if(cur == NULL)
        {
            cur = pmalloc(user->p,sizeof(_xdb));
            cur->data = data;
            cur->cache = 1;
            cur->ns = pstrdup(user->p,ns);
            cur->next = user->x_cache;
            user->x_cache = cur;
        }
    }else{
        cur = js_xdb_fetch(global, ns);
        if(cur == NULL)
        {
            cur = malloc(sizeof(_xdb));
            cur->data = data;
            cur->cache = 1;
            cur->ns = strdup(ns);
            cur->next = global;
            global = cur;
        }
    }

    if(cmd == C_GET)
    {
        log_debug(ZONE,"GET %s",ns);

        /* check cache first */
        if(cur->cache && cur->data != NULL)
        {
            return cur->data;
        }else{ /* free the stale data in the cache */
            xmlnode_free(cur->data);
            cur->data = NULL;
        }

        /* ask each xdb */
        for(i = l; i != NULL; i = i->next)
            if((*(i->c))(0,cur,user,i->arg))
                return cur->data;
    }else{
        log_debug(ZONE,"SET %s",ns);

        if(cur->data != data)
        { /* the new data is different than the one in the cache, replace */
            xmlnode_free(cur->data);
            cur->data = data;
        }

        /* ask each xdb */
        for(i = l; i != NULL; i = i->next)
            if((*(i->c))(1,cur,user,i->arg))
                break;
    }

    return NULL;
}

void js_xdb_register(xcall c, void *arg)
{
    if(c == NULL) return;

    js_xdb(C_INIT, (udata)c, (char *)arg, NULL);
}

xmlnode js_xdb_get(udata user, char *ns)
{
    if(ns == NULL) return NULL;

    return js_xdb(C_GET, user, ns, NULL);
}

void js_xdb_set(udata user, char *ns, xmlnode x)
{
    if(ns == NULL) return;

    js_xdb(C_SET, user, ns, x);
}


